#
#    Copyright 2021, NTT Communications Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import os
from signal import SIGINT
from subprocess import Popen, TimeoutExpired
from tempfile import mkstemp
from time import sleep
from typing import Optional, Tuple

import requests
from psutil import NoSuchProcess, Process

from metemcyber.core.logger import get_logger
from metemcyber.core.util import get_random_local_port, merge_config

LOGGER = get_logger(name='ngrok', file_prefix='core')

CONFIG_SECTION = 'ngrok'
DEFAULT_CONFIGS = {
    CONFIG_SECTION: {
        'ngrok_path': 'ngrok',
        'ngrok_web_port': '0',
    }
}


def ngrok_pid_filepath(app_dir: str) -> str:
    return f'{app_dir}/ngrok.pid'


class NgrokMgr():
    def __init__(self, app_dir: str, seeker_port: int = 0, config_path: Optional[str] = None
                 ) -> None:
        self.app_dir: str = app_dir
        self.seeker_port: int = seeker_port
        self.config = merge_config(config_path, DEFAULT_CONFIGS)
        self.ngrok_path: str = self.config[CONFIG_SECTION]['ngrok_path']
        pid, public_url = self.check_running()
        self.pid: int = pid
        self.public_url: Optional[str] = public_url

    def check_running(self) -> Tuple[int, Optional[str]]:  # (pid|0, public_url)
        pid_file = ngrok_pid_filepath(self.app_dir)
        try:
            with open(pid_file, 'r') as fin:
                str_data = fin.readline().strip()
            str_pid, public_url, str_saved_args = str_data.split('\t', 2)
            pid = int(str_pid)
            saved_args = str_saved_args.split('\t')
        except Exception:
            return 0, None
        try:
            proc = Process(pid)
            if proc.cmdline()[:len(saved_args)] == saved_args:
                return pid, public_url
            LOGGER.info(f'got pid({pid}) which is not a ngrok. remove defunct.')
            os.unlink(pid_file)
            return 0, None
        except NoSuchProcess:
            return 0, None

    @staticmethod
    def _get_public_url(port: int) -> str:
        query_url = f'http://localhost:{port}/api/tunnels/command_line'
        retry = 5
        last_error = None
        while retry > 0:
            try:
                sleep(1)
                response = requests.request('GET', query_url, timeout=1)
                return response.json()['public_url']
            except Exception as err:
                last_error = err
                retry -= 1
                continue
        raise Exception(f'Cannot get ngrok public_url: {last_error}')

    def _try_ngrok(self) -> Tuple[int, str]:
        web_port = int(self.config[CONFIG_SECTION]['ngrok_web_port'])
        retry = 1 if web_port > 0 else 10
        proc = None
        tmp_out, tmp_fname = mkstemp(suffix='.yml')
        try:
            os.close(tmp_out)
            args = [self.ngrok_path, 'http', '--config', tmp_fname, str(self.seeker_port)]
            while retry > 0:
                port = web_port if web_port else get_random_local_port()
                with open(tmp_fname, 'w') as fout:
                    fout.write('console_ui: false\n'
                               f'web_addr: localhost:{port}\n')
                # ngrok needs to keep running in the background
                # pylint pylint: disable=R1732
                proc = Popen(args, shell=False)
                try:
                    proc.wait(timeout=1)  # may exit with 'Address already in use'
                    del proc
                    proc = None
                    retry -= 1
                    continue
                except TimeoutExpired:  # ok
                    break
            if retry == 0:
                raise Exception('Failed launching ngrok')
            public_url = self._get_public_url(port)
            with open(ngrok_pid_filepath(self.app_dir), 'w') as fout:
                assert proc
                fout.write('\t'.join([str(proc.pid), public_url] + args))
            return proc.pid, public_url
        except Exception:
            if proc:
                proc.kill()
            raise
        finally:
            os.unlink(tmp_fname)

    def start(self) -> None:
        """launch Ngrok as another process
        """
        if self.pid:
            raise Exception(f'Already running on pid({self.pid}).')
        assert self.seeker_port > 0
        self.pid, self.public_url = self._try_ngrok()
        LOGGER.info(f'started. pid={self.pid}, public_url={self.public_url}.')

    def stop(self) -> None:
        if not self.pid:
            raise Exception('Not running')
        try:
            LOGGER.info(f'stopping process({self.pid}).')
            os.kill(self.pid, SIGINT)
            os.unlink(ngrok_pid_filepath(self.app_dir))
            self.pid = 0
        except Exception as err:
            raise Exception(f'Cannot stop ngrok(pid={self.pid})') from err



#
#    Copyright 2021, NTT Communications Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from typing import ClassVar, Dict, List, Optional, Tuple

from eth_typing import ChecksumAddress

from metemcyber.core.bc.contract import Contract, retryable_contract
from metemcyber.core.bc.util import ADDRESS0


@retryable_contract
class CTIOperator(Contract):
    contract_interface: ClassVar[Dict[int, Dict[str, str]]] = {}
    contract_id: ClassVar[str] = 'CTIOperator.sol:CTIOperator'

    def history(self, token_address: ChecksumAddress, seeker_address: Optional[ChecksumAddress],
                limit: int, offset: int = 0
                #               task_id, token,       solver,          seeker_address,  state
                ) -> List[Tuple[int, ChecksumAddress, ChecksumAddress, ChecksumAddress, int]]:
        self.log_trace()
        if self.version < 1:
            func = self.contract.functions.history(token_address, limit, offset)
        else:
            if seeker_address is None:
                seeker_address = ADDRESS0
            func = self.contract.functions.history(token_address, seeker_address, limit, offset)
        return func.call()

    def set_recipient(self):
        self.log_trace()
        func = self.contract.functions.recipientFor(self.address)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('recipientFor', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: recipientFor')
        self.log_success()

    def register_recipient(self):
        self.log_trace()
        func = self.contract.functions.registerRecipient(self.address)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('registerRecipient', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: registerRecipient')
        self.log_success()

    def register_tokens(self, token_addresses):
        self.log_trace()
        func = self.contract.functions.register(token_addresses)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('register', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: register')
        self.log_success()

    def unregister_tokens(self, token_addresses):
        self.log_trace()
        func = self.contract.functions.unregister(token_addresses)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('unregister', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: unregister')
        self.log_success()

    def accept_task(self, task_id):
        self.log_trace()
        func = self.contract.functions.accepted(task_id)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('accepted', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: accepted')
        self.log_success()

    def finish_task(self, task_id, data=''):
        self.log_trace()
        func = self.contract.functions.finish(task_id, data)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('finish', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: finish')
        self.log_success()

    def cancel_challenge(self, task_id):
        self.log_trace()
        func = self.contract.functions.cancelTask(task_id)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('cancelTask', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: cancelTask')
        self.log_success()

    def reemit_pending_tasks(self, tokens):
        self.log_trace()
        func = self.contract.functions.reemitPendingTasks(tokens)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('reemitPendingTasks', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: reemitPendingTasks')
        self.log_success()

    def check_registered(self, token_addresses):
        self.log_trace()
        func = self.contract.functions.checkRegistered(token_addresses)
        return func.call()



#
#    Copyright 2021, NTT Communications Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from typing import Dict, List, Optional, Set

from eth_typing import ChecksumAddress

from .account import Account
from .catalog import Catalog


class CatalogManager():
    def __init__(self, account: Account) -> None:
        self.account: Account = account
        #                   catalog_address  catalog_id
        self.catalogs: Dict[ChecksumAddress, int] = {}
        self.actives: Set[ChecksumAddress] = set()

    def add(self, addresses: List[ChecksumAddress], activate=False) -> None:
        for address in addresses:
            catalog = Catalog(self.account).get(address)
            self.catalogs[address] = catalog.catalog_id
            if activate:
                self.actives.add(address)

    def remove(self, addresses: List[ChecksumAddress]) -> None:
        for address in addresses:
            if address not in self.catalogs.keys():
                raise Exception(f'No such catalog: {address}')
            del self.catalogs[address]
            self.actives.discard(address)

    def activate(self, addresses: List[ChecksumAddress]) -> None:
        for address in addresses:
            if address not in self.catalogs.keys():
                raise Exception(f'No such catalog: {address}')
            self.actives.add(address)

    def deactivate(self, addresses: List[ChecksumAddress]) -> None:
        for address in addresses:
            if address not in self.catalogs.keys():
                raise Exception(f'No such catalog: {address}')
            self.actives.discard(address)

    def _catalogs(self, active: Optional[bool]) -> Dict[ChecksumAddress, int]:
        return {addr: cid for addr, cid in self.catalogs.items()
                if active in (None, addr in self.actives)}

    @property
    def active_catalogs(self) -> Dict[ChecksumAddress, int]:
        return self._catalogs(active=True)

    @property
    def reserved_catalogs(self) -> Dict[ChecksumAddress, int]:
        return self._catalogs(active=False)

    @property
    def all_catalogs(self) -> Dict[ChecksumAddress, int]:
        return self._catalogs(active=None)

    def id2address(self, catalog_id: int) -> ChecksumAddress:
        tmp = [caddr for caddr, cid in self.catalogs.items()
               if cid == catalog_id]
        if not tmp:
            raise Exception('No such catalog id: ' + str(catalog_id))
        assert len(tmp) == 1
        return tmp[0]

    def address2id(self, address: ChecksumAddress) -> int:
        return self.catalogs[address]



#
#    Copyright 2021, NTT Communications Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import logging
import os
import sys
from threading import Thread
from time import sleep
from typing import Callable, Optional, Tuple

from flask import Flask, request
from werkzeug.datastructures import EnvironHeaders

from .util import get_random_local_port

RANDOM_PORT_RETRY_MAX = 10

os.environ['WERKZEUG_RUN_MAIN'] = 'true'  # eliminate werkzeug server banner on boot
logging.getLogger('werkzeug').disabled = True


class WebhookReceiver():
    def __init__(self, address: str, port: int,
                 callback: Optional[Callable[[EnvironHeaders, str], None]] = None) -> None:
        assert port >= 0
        self.address: str = address
        self.port: int = port
        self.callback: Callable[[EnvironHeaders, str],
                                None] = callback if callback else self.resolve_request
        self.thread: Optional[Thread] = None

    def start(self) -> Tuple[str, int]:
        if self.thread:
            raise Exception('Already running')
        self.thread = Thread(target=self.boot_server, daemon=True)
        self.thread.start()
        sleep(1)
        assert self.thread
        assert self.thread.is_alive()
        return self.address, self.port

    def boot_server(self) -> None:
        app = Flask(self.__class__.__name__)
        app.add_url_rule('/', 'post_callback', self._data_dispatcher, methods=['POST'])
        if self.port > 0:
            app.run(self.address, self.port)
            return
        retry = RANDOM_PORT_RETRY_MAX
        while True:
            try:
                self.port = get_random_local_port()
                app.run(self.address, self.port)
            except OSError as err:
                if err.errno == 98:  # Address already in use
                    retry -= 1
                    if retry > 0:
                        continue
                    raise Exception('Cannot get unused port') from err
            break
        self.thread = None

    def _data_dispatcher(self) -> str:
        headers = request.headers
        body = request.get_data(cache=False, as_text=True)  # XXX: should decode as text?
        Thread(target=self.callback, args=[headers, body], daemon=True).start()
        return 'ok'

    def resolve_request(self, headers: EnvironHeaders, body: str) -> None:
        print(self)
        print(headers)
        print('--')
        print(body)


def main(listen_address: str, listen_port: str):
    try:
        webhook = WebhookReceiver(listen_address, int(listen_port))
        address, port = webhook.start()
        pid = os.getpid()
        print(f'{pid}\t{address}\t{port}')
        assert webhook.thread
        webhook.thread.join()
    except KeyboardInterrupt:
        print('caught SIGINT.')


if __name__ == '__main__':
    main(sys.argv[1], sys.argv[2])



#
#    Copyright 2021, NTT Communications Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from __future__ import annotations

from typing import Dict, List, Optional

from eth_typing import ChecksumAddress

from .account import Account
from .catalog import Catalog
from .cti_broker import CTIBroker
from .cti_token import CTIToken
from .token import Token

PTS_RATE: int = 10**18  # FIXME: should import from somewhere


class Broker():
    #            {broker:              {catalog:             {token:           amount}}}
    amounts: Dict[ChecksumAddress, Dict[ChecksumAddress, Dict[ChecksumAddress, int]]] = {}

    def __init__(self, account: Account) -> None:
        self.account: Account = account
        self.address: Optional[ChecksumAddress] = None

    def get(self, address: ChecksumAddress) -> Broker:
        self.address = address
        return self

    def new(self) -> Broker:
        cti_broker = CTIBroker(self.account).new()
        return self.get(cti_broker.address)

    def uncache(self, catalog: Optional[ChecksumAddress] = None,
                token: Optional[ChecksumAddress] = None,
                entire: bool = False) -> None:
        if entire:
            del Broker.amounts
            Broker.amounts = {}
            return
        assert self.address
        if self.address not in Broker.amounts.keys():
            return
        if catalog is None:
            Broker.amounts.pop(self.address, None)
            return
        if catalog not in Broker.amounts[self.address].keys():
            return
        if token is None:
            Broker.amounts[self.address].pop(catalog, None)
            return
        Broker.amounts[self.address][catalog].pop(token, None)

    def _fill_amounts(self, catalog: ChecksumAddress) -> None:
        assert self.address
        if self.address not in Broker.amounts.keys():
            Broker.amounts[self.address] = {}
        bmap: dict = Broker.amounts[self.address]
        if catalog not in bmap.keys():
            bmap[catalog] = {}
        cmap: dict = bmap[catalog]
        queries = list(Catalog(self.account).get(catalog).tokens.keys() - cmap.keys())
        if len(queries) == 0:  # already filled
            return
        cti_broker = CTIBroker(self.account).get(self.address)
        amounts = cti_broker.get_amounts(catalog, queries)
        for idx, token in enumerate(queries):
            cmap[token] = amounts[idx]

    def get_amount(self, catalog: ChecksumAddress, token: ChecksumAddress) -> int:
        assert self.address
        self._fill_amounts(catalog)
        try:
            return Broker.amounts[self.address][catalog][token]
        except KeyError as err:
            raise Exception('No such token or catalog or broker') from err

    def get_amounts(self, catalog: ChecksumAddress, tokens: List[ChecksumAddress]) -> List[int]:
        assert self.address
        self._fill_amounts(catalog)
        try:
            return [Broker.amounts[self.address][catalog][token] for token in tokens]
        except KeyError as err:
            raise Exception('No such token or catalog or broker') from err

    def consign(self, catalog: ChecksumAddress, token: ChecksumAddress,
                amount: int) -> None:
        assert self.address
        cti_token = CTIToken(self.account).get(token)
        cti_broker = CTIBroker(self.account).get(self.address)
        cti_token.authorize_operator(self.address)
        try:
            cti_broker.consign_token(catalog, token, amount)
            self.uncache(catalog, token)
            Token(self.account).get(token).uncache()  # FIXME: should I?
        finally:
            cti_token.revoke_operator(self.address)

    def takeback(self, catalog: ChecksumAddress, token: ChecksumAddress,
                 amount: int) -> None:
        assert self.address
        cti_broker = CTIBroker(self.account).get(self.address)
        cti_broker.takeback_token(catalog, token, amount)
        self.uncache(catalog, token)
        Token(self.account).get(token).uncache()  # FIXME: should I?

    def buy(self, catalog: ChecksumAddress, token: ChecksumAddress,
            price: int, allow_cheaper: bool = False) -> None:
        assert self.address
        wei = price * PTS_RATE
        cti_broker = CTIBroker(self.account).get(self.address)
        cti_broker.buy_token(catalog, token, wei, allow_cheaper)
        self.uncache(catalog, token)
        Token(self.account).get(token).uncache()  # FIXME: should I?



#
#    Copyright 2021, NTT Communications Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from web3 import Web3
from web3.exceptions import ExtraDataLengthError
from web3.middleware import construct_sign_and_send_raw_middleware, geth_poa_middleware
from web3.providers.rpc import HTTPProvider


class Ether:
    def __init__(self, endpoint):
        self.web3 = Web3(HTTPProvider(endpoint))
        if self.web3.isConnected():
            try:
                self.web3.eth.getBlock('latest')
            except ExtraDataLengthError:
                self.web3.middleware_onion.inject(geth_poa_middleware, layer=0)

    def signature(self, private_key):

        # 署名付きトランザクションの準備とcoinbaseアカウントの作成
        cssrm = construct_sign_and_send_raw_middleware(private_key)
        acc = self.web3.eth.account.from_key(private_key)

        # プライベート鍵の変数は今後使わないので明示的に削除
        del private_key

        # ミドルウェアに署名付きトランザクションを設定
        # See
        # https://github.com/ethereum/web3.py/blob/master/web3/datastructures.py
        # pylint: disable=protected-access
        if 'sign_and_send_raw' in self.web3.middleware_onion._queue:
            self.web3.middleware_onion.replace('sign_and_send_raw', cssrm)
        else:
            self.web3.middleware_onion.add(cssrm, 'sign_and_send_raw')

        # coinbaseとしてアカウントを設定
        self.web3.eth.defaultAccount = acc.address

    def web3_with_signature(self, private_key):
        self.signature(private_key)
        return self.web3



#
#    Copyright 2021, NTT Communications Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

class Wallet:
    def __init__(self, web3, eoa):
        self.web3 = web3
        self.eoa = eoa
        self.__balance = self._get_balance()

    def _get_balance(self):
        return self.web3.eth.getBalance(self.eoa)

    @property
    def balance(self):
        self.__balance = self._get_balance()
        return self.__balance



#
#    Copyright 2021, NTT Communications Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import json
import os
import sys
from signal import SIGINT
from subprocess import Popen
from time import sleep
from typing import List, Optional, Tuple, cast
from urllib.request import Request, urlopen

from eth_typing import ChecksumAddress
from psutil import NoSuchProcess, Process
from werkzeug.datastructures import EnvironHeaders

from metemcyber.cli.constants import APP_DIR
from metemcyber.core.bc.account import Account
from metemcyber.core.bc.ether import Ether
from metemcyber.core.bc.operator import Operator
from metemcyber.core.bc.util import verify_message
from metemcyber.core.logger import get_logger
from metemcyber.core.solver import SIGNATURE_HEADER
from metemcyber.core.util import merge_config
from metemcyber.core.webhook import WebhookReceiver

LOGGER = get_logger(name='seeker', file_prefix='core')
TTY_FILEPATH = f'{APP_DIR}/.tty4seeker.lnk'
LIMIT_ATONCE = 16

CONFIG_SECTION = 'seeker'
DEFAULT_CONFIGS = {
    CONFIG_SECTION: {
        'downloaded_cti_path': f'{APP_DIR}/workspace/download',
        'listen_address': '127.0.0.1',
        'listen_port': '0',
        'ngrok': '0',
    }
}


def tty_message(msg: str) -> None:
    try:
        with open(TTY_FILEPATH, 'w') as tty:
            if not tty.isatty:
                return
            tty.write(msg + '\n')
    except Exception:
        pass


def seeker_pid_filepath(app_dir: str) -> str:
    return f'{app_dir}/seeker.pid'


def download_json(download_url: str, token_address: ChecksumAddress, dir_path: str) -> None:
    try:
        request = Request(download_url, method='GET')
        with urlopen(request) as response:
            rdata = response.read()
            if isinstance(rdata, bytes):
                rdata = rdata.decode()
    except Exception as err:
        LOGGER.exception(err)
        msg = f'Failed download from {download_url}.'
        LOGGER.error(msg)
        tty_message(msg)
        return

    try:
        jdata = json.loads(rdata)
        title = jdata['Event']['info']
    except Exception:
        title = '(cannot decode title)'
    msg = f'Downloaded data. title: "{title}".'
    LOGGER.info(msg)
    tty_message(msg)

    try:
        if not os.path.isdir(dir_path):
            os.makedirs(dir_path)
        filepath = f'{dir_path}/{token_address}.json'
        with open(filepath, 'w') as fout:
            json.dump(jdata, fout, ensure_ascii=False, indent=2)
        msg = f'Saved downloaded data in {filepath}.'
        LOGGER.info(msg)
        tty_message(msg)
    except Exception as err:
        LOGGER.exception(err)
        msg = f'Saving downloaded data failed: {err}'
        LOGGER.error(msg)
        tty_message(msg)

    tty_message('(type CTRL-C to quit monitoring)')


class Resolver(WebhookReceiver):
    def __init__(self, listen_address: str, listen_port: int,
                 download_path: str, endpoint_url: str, operator_address: ChecksumAddress):
        super().__init__(listen_address, listen_port)
        self.download_path = download_path
        self.endpoint_url = endpoint_url
        self.operator_address = operator_address

    def resolve_request(self, headers: EnvironHeaders, body: str):
        sign = headers.get(SIGNATURE_HEADER)
        if sign is None:
            msg = f'Received request without signature.'
            LOGGER.error(msg)
            tty_message(msg)
            return
        LOGGER.debug(f'Received request. headers={headers}, body={body}.')
        tty_message('Incoming data...')
        try:
            jdata = json.loads(body)
            if jdata['from'] != verify_message(body, sign):
                raise Exception('Signer mismatch.')
            tty_message(f'Data sender: {jdata["from"]}.')
            operator = Operator(Account(Ether(self.endpoint_url))).get(self.operator_address)
            task = None
            offset = 0
            while True:
                tasks = operator.history(jdata['token_address'], None, LIMIT_ATONCE, offset)
                tmp = [item for item in tasks if item[0] == int(jdata['task_id'])]
                if tmp:
                    task = tmp[0]
                    break
                if len(tasks) < LIMIT_ATONCE:
                    break
                offset += LIMIT_ATONCE

            if task is None:
                raise Exception(f'No such task id: {jdata["task_id"]}')
            _, t_addr, t_solver, _, _ = task
            if t_addr != jdata['token_address'] or t_solver != jdata['from']:
                raise Exception('Task info mismatch')
        except KeyError as err:
            msg = f'Missing parameter: {err}'
            LOGGER.error(msg)
            tty_message(msg)
            return
        except Exception as err:
            LOGGER.exception(err)
            msg = f'Failed verifying data from solver: {err}'
            LOGGER.error(msg)
            tty_message(msg)
            return

        msg = f'Trying download_url for task {jdata["task_id"]} - ' + \
              f'token({jdata["token_address"]}): {jdata["download_url"]}'
        LOGGER.info(msg)
        tty_message(msg)
        download_json(jdata['download_url'], jdata['token_address'], self.download_path)


class Seeker():
    cmd_args_base = ['python3', __file__]

    @property
    def cmd_args(self) -> List[str]:
        args = Seeker.cmd_args_base + [
            self.config[CONFIG_SECTION]['listen_address'],
            self.config[CONFIG_SECTION]['listen_port'],
            self.app_dir,
            self.operator_address,
            self.endpoint_url,
        ]
        if self.config_path:
            args += [self.config_path]
        return args

    def __init__(self, app_dir: str, operator_address: ChecksumAddress, endpoint_url: str = '',
                 config_path: Optional[str] = None
                 ) -> None:
        self.app_dir = app_dir
        self.config_path = config_path
        self.config = merge_config(config_path, DEFAULT_CONFIGS)
        pid, address, port = self.check_running()
        self.pid: int = pid
        self.address: Optional[str] = address
        self.port: int = port
        self.operator_address = operator_address
        self.endpoint_url = endpoint_url
        if operator_address and endpoint_url:
            operator = Operator(Account(Ether(endpoint_url))).get(operator_address)
            if operator.version < 1:
                raise Exception(
                    f'Operator({operator_address}) is version {operator.version} and '
                    'not support anonymous access.')

    #                               (pid|0, listen_address, listen_port)
    def check_running(self) -> Tuple[int, Optional[str], int]:
        try:
            with open(seeker_pid_filepath(self.app_dir), 'r') as fin:
                str_data = fin.readline().strip()
            str_pid, address, str_port = str_data.split('\t', 2)
            pid = int(str_pid)
        except Exception:
            return 0, None, 0
        try:
            proc = Process(pid)
            running_seeker = proc.cmdline()[:len(Seeker.cmd_args_base)]
            if len(running_seeker) > 1 and len(Seeker.cmd_args_base) > 1:
                # Python path locations are not always the same, so compare with arguments.
                if running_seeker[1] == Seeker.cmd_args_base[1]:
                    return pid, address, int(str_port)
            # found pid, but it's not a seeker. remove defunct data.
            LOGGER.info(f'got pid({pid}) which is not a seeker. remove defunct.')
            os.unlink(seeker_pid_filepath(self.app_dir))
            return 0, None, 0
        except NoSuchProcess:
            return 0, None, 0

    def start(self) -> None:
        """launch Resolver by calling main() as another process
        """
        if self.pid:
            raise Exception(f'Already running on pid({self.pid}).')
        # Seeker needs to keep running in the background.
        # pylint pylint: disable=R1732
        proc = Popen(self.cmd_args, shell=False)
        for _cnt in range(5):
            sleep(1)
            self.pid, self.address, self.port = self.check_running()
            if self.pid:
                LOGGER.info(f'started. pid={self.pid}, address={self.address}, port={self.port}.')
                return
        LOGGER.error('Cannot start webhook.')
        proc.kill()
        raise Exception('Cannot start webhook')

    def stop(self) -> None:
        pid, _address, _port = self.check_running()
        if not pid:
            raise Exception('Not running')
        try:
            LOGGER.info(f'stopping process({pid}).')
            os.kill(pid, SIGINT)
        except Exception as err:
            raise Exception(f'Cannot stop webhook(pid={pid})') from err


def main(argv: List[str]):
    listen_address, listen_port, app_dir, operator_address, endpoint_url = argv[:5]
    config_path = argv[5] if len(argv) > 5 else None
    pid_file = seeker_pid_filepath(app_dir)
    try:
        config = merge_config(config_path, DEFAULT_CONFIGS)
        resolver = Resolver(listen_address, int(listen_port),
                            config[CONFIG_SECTION]['downloaded_cti_path'],
                            endpoint_url,
                            cast(ChecksumAddress, operator_address))
        address, port = resolver.start()
        pid = os.getpid()
        with open(pid_file, 'w') as fout:
            fout.write(f'{pid}\t{address}\t{port}\n')
        assert resolver.thread
        resolver.thread.join()
    except KeyboardInterrupt:
        LOGGER.info('caught SIGINT.')
    finally:
        if os.path.exists(pid_file):
            os.unlink(pid_file)


if __name__ == '__main__':
    if len(sys.argv) < 5:
        raise Exception('Not enough arguments')
    main(sys.argv[1:])



#
#    Copyright 2021, NTT Communications Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from typing import ClassVar, Dict, List, Tuple

from eth_typing import ChecksumAddress
from web3 import Web3

from metemcyber.core.bc.contract import Contract, ContractVersionError, retryable_contract


@retryable_contract
class CTIToken(Contract):
    contract_interface: ClassVar[Dict[int, Dict[str, str]]] = {}
    contract_id: ClassVar[str] = 'CTIToken.sol:CTIToken'

    @property
    def publisher(self) -> ChecksumAddress:
        self.log_trace()
        func = self.contract.functions.publisher()
        return func.call()

    def balance_of(self, account_id):
        self.log_trace()
        func = self.contract.functions.balanceOf(account_id)
        return func.call()

    def mint(self, dest: ChecksumAddress, amount: int,
             user_data: str = '', operator_data: str = ''):
        self.log_trace()
        bdata_user = Web3.toBytes(text=user_data)
        bdata_operator = Web3.toBytes(text=operator_data)
        func = self.contract.functions.mint(dest, amount, bdata_user, bdata_operator)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: mint')
        self.log_success()

    def send_token(self, dest, amount=1, data=''):
        self.log_trace()
        bdata = Web3.toBytes(text=data)
        func = self.contract.functions.send(dest, amount, bdata)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('send', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: send')
        self.log_success()

    def burn_token(self, amount, data=''):
        self.log_trace()
        bdata = Web3.toBytes(text=data)
        func = self.contract.functions.burn(amount, bdata)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('burn', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: burn')
        self.log_success()

    def authorize_operator(self, operator):
        self.log_trace()
        func = self.contract.functions.authorizeOperator(operator)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('authorizeOperator', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: authorizeOperator')
        self.log_success()

    def revoke_operator(self, operator):
        self.log_trace()
        func = self.contract.functions.revokeOperator(operator)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('revokeOperator', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: revokeOperator')
        self.log_success()

    @property
    def editable(self) -> bool:
        if self.version < 1:
            raise ContractVersionError('Not supported (too old contract version)')
        self.log_trace()
        func = self.contract.functions.editable()
        return func.call()

    def set_editable(self, editable: bool):
        if self.version < 1:
            raise ContractVersionError('Not supported (too old contract version)')
        self.log_trace()
        func = self.contract.functions.setEditable(editable)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('setEditable', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: setEditable')
        self.log_success()

    def add_candidates(self, candidates: List[str]):
        if self.version < 1:
            raise ContractVersionError('Not supported (too old contract version)')
        self.log_trace()
        func = self.contract.functions.addCandidates(candidates)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('addCandidates', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: addCandidates')
        self.log_success()

    def remove_candidates(self, indexes: List[int]):
        if self.version < 1:
            raise ContractVersionError('Not supported (too old contract version)')
        self.log_trace()
        func = self.contract.functions.removeCandidates(indexes)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('removeCandidatas', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: removeCandidates')
        self.log_success()

    def list_candidates(self) -> List[Tuple[int, int, str]]:  # [(index, score, desc), ...]
        if self.version < 1:
            raise ContractVersionError('Not supported (too old contract version)')
        self.log_trace()
        func = self.contract.functions.listCandidates()
        return func.call()

    def vote(self, index: int, amount: int):
        if self.version < 1:
            raise ContractVersionError('Not supported (too old contract version)')
        self.log_trace()
        func = self.contract.functions.vote(index, amount)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('vote', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('Transaction failed: vote')
        self.log_success()



#
#    Copyright 2021, NTT Communications Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import logging
import logging.handlers
from pathlib import Path
from typing import List

from typer import get_app_dir


def get_logger(name: str, app_dir: str = "", file_prefix: str = ""):
    return MetemcyberLogger(name=name, app_dir=app_dir,
                            file_prefix=file_prefix).logger


class MetemcyberLogger():
    """
    Class for logging metemcyber Clis
    """
    created_loggers: List[str] = []

    def __init__(self, name: str, app_dir: str = "",
                 file_prefix: str = "") -> None:
        """Initilizates a Logger for Metemcyber Clis.

        :param name: Strings of logger name.
        :param log_file: Strings of log file name.
        :param log_level: Strings of Log level.
        """

        # The default location
        app_default_dir = get_app_dir("metemcyber")
        # Max bytes of log file
        max_bytes: int = 32768000
        # The number of backup of log file
        backup_num: int = 3

        # If already registered, return logger
        if name in MetemcyberLogger.created_loggers:
            self.logger: logging.Logger = logging.getLogger(name)
        else:
            # Use a directory of this app
            if not app_dir:
                app_dir = app_default_dir
            log_path: Path = Path(app_dir) / "logs"
            log_path.mkdir(parents=True, exist_ok=True)

            # Setup Logger
            self.logger = logging.getLogger(name)
            self.logger.setLevel(logging.DEBUG)

            user_formatter = UserFormatter(
                '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            formatter = logging.Formatter(
                '%(asctime)s - %(name)s - %(levelname)s - %(message)s')

            # Set log file name
            if file_prefix != "":
                error_filename = f"{file_prefix}-error.log"
                debug_filename = f"{file_prefix}-debug.log"
            else:
                error_filename = "error.log"
                debug_filename = "debug.log"

            # Setup error log file handler
            rth_error = logging.handlers.RotatingFileHandler(
                filename=log_path / error_filename,
                maxBytes=max_bytes,
                backupCount=backup_num,
            )
            rth_error.setLevel(logging.WARNING)
            rth_error.setFormatter(user_formatter)
            self.logger.addHandler(rth_error)

            # Setup debug log file handler
            rth_debug = logging.handlers.RotatingFileHandler(
                filename=log_path / debug_filename,
                maxBytes=max_bytes,
                backupCount=backup_num,
            )
            rth_debug.setLevel(logging.DEBUG)
            rth_debug.setFormatter(formatter)
            self.logger.addHandler(rth_debug)

            MetemcyberLogger.created_loggers.append(name)


class UserFormatter(logging.Formatter):
    def formatException(self, _):
        # TODO: StackTrace発生時、ユーザーに通知したいコメントを書く
        return None



#
#    Copyright 2021, NTT Communications Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from typing import ClassVar, Dict

from metemcyber.core.bc.contract import Contract, retryable_contract


@retryable_contract
class CTIBroker(Contract):
    contract_interface: ClassVar[Dict[int, Dict[str, str]]] = {}
    contract_id: ClassVar[str] = 'CTIBroker.sol:CTIBroker'

    def consign_token(self, catalog, token, amount):
        self.log_trace()
        func = self.contract.functions.consignToken(catalog, token, amount)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('consignToken', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('consignToken: transaction failed')
        self.log_success()

    def takeback_token(self, catalog, token, amount):
        self.log_trace()
        func = self.contract.functions.takebackToken(catalog, token, amount)
        tx_hash = func.transact()
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('takebackToken', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('takebackToken: transaction failed')
        self.log_success()

    def buy_token(self, catalog, token, wei, allow_cheaper=False):
        self.log_trace()
        func = self.contract.functions.buyToken(catalog, token, allow_cheaper)
        tx_hash = func.transact({'value': wei})
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        self.gaslog('buyToken', tx_receipt)
        if tx_receipt['status'] != 1:
            raise ValueError('buyToken: transaction failed')
        self.log_success()

    def get_amounts(self, catalog, tokens):
        self.log_trace()
        func = self.contract.functions.getAmounts(catalog, tokens)
        return func.call()
